import os
import readline
import itertools


def no_arg_completion(text):
    return []


def commit_arg_completion(text):
    return ["confirmed"]


def fix_completion_dirname(name):
    if os.path.isdir(name) and name[-1] != "/":
        return name + "/"
    else:
        return name


def filename_arg_completion(text):
    dir = os.path.dirname(text)
    rdir = "." if dir == "" else dir
    return [fix_completion_dirname(os.path.join(dir, name)) for name in os.listdir(rdir)]


def validate_arg_completion(text):
    return ["candidate"] + filename_arg_completion(text)


class NCCompleter(object):
    def __init__(self, operations, option_args):
        self.operations = operations
        self.option_args = option_args

    def __call__(self, text, state):
        return self.complete_with(text, state, readline.get_line_buffer())

    def complete_command_argument(self, operation, text):
        if operation.name in ["kill_session", "get_schema", "create_subscription", "sleep"]:
            return []
        if operation.name == "commit":
            return ["confirmed"]

    def lookup_option(self, option_text):
        for (opts, descriptor) in self.option_args.values():
            if option_text in opts:
                return descriptor
        return None

    def complete_command_options(self, text, line_args):
        opname = line_args[0]
        try:
            operation = self.operations[opname]
        except KeyError:
            return []
        cmdargs = []
        word_count = len(line_args)
        if text == "":
            # the user pressed TAB for a new command option
            word_count += 1
        if word_count == 2:
            # command argument (filename for edit-config, datastore for validate, ...)
            cmdargs = operation.arg_completion(text)
            if operation.nargs == 1:
                return cmdargs
        if word_count > 2:
            opt_descriptor = self.lookup_option(line_args[word_count - 2])
            if opt_descriptor is not None:
                if "choices" in opt_descriptor:
                    return opt_descriptor["choices"]
                elif opt_descriptor["dest"] == "db":
                    return ["running", "candidate"]
                return []  # no completion for timeout or xpath
        opts = [self.option_args[option][0] for option in operation.command_opts]
        return cmdargs + list(itertools.chain(*opts))

    def complete_with(self, text, state, line):
        line_args = line.split()
        if len(line_args) > 1 or len(line_args) == 1 and text == "":
            variants = self.complete_command_options(text, line_args)
        else:
            variants = self.operations.keys()
        permitted = sorted(var for var in variants if var.startswith(text))
        if len(permitted) <= state:
            return None
        return permitted[state]



"""NETCONF over plain TCP implementation. Derives heavily from SSH to
reuse in/out processing methods."""

from ncclient import transport
import socket
from contextlib import contextmanager
import os


class SSHSession(transport.SSHSession):
    "SSHSession extension capable of saving the raw data."
    def __init__(self, device_handler, raw_file):
        self.raw_file = raw_file
        super(SSHSession, self).__init__(device_handler)
        self.last_pos = 0

    def _parse10(self):
        with self.raw_processing(self._parsing_pos10):
            super(SSHSession, self)._parse10()

    def _parse11(self):
        with self.raw_processing(self._parsing_pos11):
            super(SSHSession, self)._parse11()

    @contextmanager
    def raw_processing(self, position):
        if self.raw_file is not None:
            self.raw_file.write(self._buffer.getvalue()[self.last_pos:].decode("utf8"))
        try:
            yield
        finally:
            self.last_pos = self._buffer.tell()


class TCPChannel(object):
    def __init__(self, host, port):
        for res in socket.getaddrinfo(host, port,
                                      socket.AF_UNSPEC, socket.SOCK_STREAM):
            af, socktype, proto, canonname, sa = res
            try:
                sock = socket.socket(af, socktype, proto)
            except socket.error:
                sock = None
                continue
            try:
                sock.connect(sa)
            except socket.error:
                sock.close()
                sock = None
                continue
            break
        if sock is None:
            raise Exception("Could not connect to the host")
        self.sock = sock
        self.active = True

    def fileno(self):
        return self.sock.fileno()

    def close(self):
        self.sock.close()
        self.active = False

    def is_active(self):
        return self.active

    def recv(self, *args):
        return self.sock.recv(*args)

    def send(self, data):
        return self.sock.send(data.encode('UTF-8'))

    def send_ready(self):
        return True


class TCPSession(SSHSession):
    """TCP session as an extension of SSH session - parsing is the same,
but connect and authentication differs greatly.

    """
    def connect(self, host, port=2023, username=None, password=None, **ignored):
        self._channel = TCPChannel(host, port)
        sockname = self._channel.sock.getsockname()
        self._channel.send("[%s;%s;tcp;%d;%d;%s;%s;%s;]\n" %
                           (username, sockname[0], os.getuid(), os.getgid(), "",
                            os.getenv("HOME", "/tmp"), ""))
        self._connected = True
        self._transport = self._channel
        self._post_connect()



import _confd
from _confd import XmlTag, Value, TagValue
from _confd import dp
import socket
from notifier_ns import ns
from datetime import datetime as dt


def init(dname="ncpy"):
    global ctrsock, dx, sock  # must not be deallocated
    ctrsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    dx = dp.init_daemon(dname)
    dp.connect(dx, ctrsock, dp.CONTROL_SOCKET, '127.0.0.1', 4565)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    dp.connect(dx, sock, dp.WORKER_SOCKET, '127.0.0.1', 4565)
    stream = dp.register_notification_stream(dx, object(), sock, 'test-stream')
    dp.register_done(dx)
    return stream


def notify_value(stream, value):
    tagval = [TagValue(XmlTag(ns.hash, ns.notifier_pythonnotif),
                       Value((ns.notifier_pythonnotif, ns.hash), _confd.C_XMLBEGIN)),
              TagValue(XmlTag(ns.hash, ns.notifier_whatsup),
                       Value(value)),
              TagValue(XmlTag(ns.hash, ns.notifier_pythonnotif),
                       Value((ns.notifier_pythonnotif, ns.hash), _confd.C_XMLEND))]
    dp.notification_send(stream, _confd.DateTime(*dt.now().timetuple()), tagval)



"Abstractions of operations that can be invoked on a Netconf server."

import time
import sys
import re
from lxml import etree
import ncclient
from ncclient.xml_ import new_ele, sub_ele, qualify, validated_element
from ncclient import manager
from ncclient.operations import util
from ncclient.transport import SessionListener
from ncclient.operations.rpc import RPC
from ncclient.operations import retrieve
from . import completions

NOTIFICATION_NS = "urn:ietf:params:xml:ns:netconf:notification:1.0"
TAILF_AAA_NS = "http://tail-f.com/ns/aaa/1.1"
IETF_NACM_NS = "urn:ietf:params:xml:ns:yang:ietf-netconf-acm"
DEFAULTS_NS = "urn:ietf:params:xml:ns:yang:ietf-netconf-with-defaults"
INACTIVE_NS = "http://tail-f.com/ns/netconf/inactive/1.0"


class ExecuteRpc(RPC):
    "Custom general-purpose RPC"
    def request(self, rpc):
        return self._request(rpc)


def extend_get_node(rpc, node, filter=None, defaults=None, inactive=None):
    if filter is not None:
        node.append(util.build_filter(filter, rpc._assert))
    if defaults is not None:
        etree.SubElement(node, qualify("with-defaults", DEFAULTS_NS)).text = defaults
    if inactive is not None:
        etree.SubElement(node, qualify("with-inactive", INACTIVE_NS))
    return node


class GetConfigRpc(retrieve.GetConfig):
    """The *get-config* RPC, with support for with-defaults."""

    def request(self, source, filter=None, defaults=None, nsmap={}):
        node = new_ele("get-config", nsmap=nsmap)
        node.append(util.datastore_or_url("source", source, self._assert))
        return self._request(extend_get_node(self, node, filter, defaults))


class GetRpc(retrieve.Get):
    """The *get* RPC, with support for with-defaults."""

    def request(self, filter=None, defaults=None, nsmap={}):
        node = new_ele("get", nsmap=nsmap)
        return self._request(extend_get_node(self, node, filter, defaults))


class ValidateRpc(RPC):
    "*validate* RPC, with support for config data"
    # The Validate implementation in ncclient is invalid now (Jun 2016)

    DEPENDS = [":validate"]

    def request(self, source):
        """Validate the contents of the specified configuration.

        *source* is the name of the configuration datastore being
         validated or `config` element containing the configuration
         subtree to be validated

        :seealso: :ref:`srctarget_params`
        """
        node = new_ele("validate")
        if type(source) is str:
            src = util.datastore_or_url("source", source, self._assert)
        else:
            validated_element(source, ("config", qualify("config")))
            src = new_ele("source")
            src.append(source)
        node.append(src)
        return self._request(node)


manager.VENDOR_OPERATIONS.update(rpc=ExecuteRpc, get_config_ncc=GetConfigRpc, get_ncc=GetRpc,
                                 validate_ncc=ValidateRpc)


class Operation(object):
    name = None
    option = None
    help = None
    nargs = 0
    dest = None
    command_opts = []
    choices = None
    arg_completion = staticmethod(completions.no_arg_completion)

    def invoke(self, mc, ns):
        pass


class Hello(Operation):
    name = option = "hello"
    help = "Connect to the server and print its capabilities"

    def invoke(self, mc, *ignored):
        caps = list(mc.server_capabilities)
        hello = new_ele("hello")
        xcaps = sub_ele(hello, "capabilities")
        for cap in caps:
            sub_ele(xcaps, "capability").text = cap
        return hello


class GetOperation(Operation):
    command_opts = ["style", "wdefaults", "xpath"]
    nargs = "?"

    def invoke(self, mc, ns, path=None):
        replydata = self._invoke(mc, ns, path)
        if "noaaa" in ns.style:
            aaa = replydata.find(qualify("aaa", TAILF_AAA_NS))
            if aaa is not None:
                replydata.remove(aaa)
            nacm = replydata.find(qualify("nacm", IETF_NACM_NS))
            if nacm is not None:
                replydata.remove(nacm)
        return replydata


class Get(GetOperation):
    name = option = "get"
    help = "Takes an optional -x argument for XPath filtering"

    def _invoke(self, mc, ns, path=None):
        reply = mc.get_ncc(xpath_filter(ns, path), ns.wdefaults, nsmap(ns))
        return reply.data


class GetConfig(GetOperation):
    name = "get_config"
    option = "get-config"
    help = "Takes an optional --db argument, default is 'running'" \
           ", and an optional -x argument for XPath filtering"
    command_opts = GetOperation.command_opts + ["db"]

    def _invoke(self, mc, ns, path=None):
        reply = mc.get_config_ncc(ns.db, xpath_filter(ns, path), ns.wdefaults, nsmap(ns))
        return reply.data


class DiscardChanges(Operation):
    name = "discard_changes"
    option = "discard-changes"

    def invoke(self, mc, ns):
        reply = mc.discard_changes()
        return reply._root[0]


class Commit(Operation):
    name = "commit"
    option = "commit"
    nargs = "?"
    choices = ['confirmed']
    command_opts = ["timeout"]
    arg_completion = staticmethod(completions.commit_arg_completion)

    def invoke(self, mc, ns, confirmed=False):
        reply = mc.commit(confirmed, str(ns.timeout))
        return reply._root[0]


class KillSession(Operation):
    name = "kill_session"
    option = "kill-session"
    help = "Takes a session-id as argument"
    nargs = 1
    dest = "session_id"

    def invoke(self, mc, ns, session):
        reply = mc.kill_session(session)
        return reply._root[0]


class Validate(Operation):
    name = option = "validate"
    help = ("Takes either 'candidate', or a filename or '-' as an optinal"
            " argument, the default is '-'.  If it is 'candidate', the request for"
            " validation of the candidate datastore is sent, otherwise the contents"
            " of the file is sent as configuration to be validated.")
    nargs = "?"
    arg_completion = staticmethod(completions.validate_arg_completion)

    def invoke(self, mc, ns, src="-"):
        if src == "candidate":
            source = src
        else:
            source = etree.parse(sys.stdin if src == "-"
                                 else open(src, "r")).getroot()
        reply = mc.validate_ncc(source)
        return reply._root[0]


class CopyRunningToStartup(Operation):
    option = "copy-running-to-startup"
    name = "copy_running_to_startup"

    def invoke(self, mc, ns):
        reply = mc.copy_config("running", "startup")
        return reply._root[0]


class CopyConfig(Operation):
    option = "copy-config"
    name = "copy_config"
    nargs = "?"
    arg_completion = staticmethod(completions.filename_arg_completion)
    help = ("Takes a filename or '-' as an optional"
            " argument. The contents of the file"
            " is data for a single NETCONF copy-config operation"
            " (put into the <config> XML element)."
            " Takes an optional --db argument, default is 'running'.")
    command_opts = ["db"]

    def invoke(self, mc, ns, filename="-"):
        data = etree.parse(sys.stdin if filename == "-" else open(filename, "r"))
        copy = new_ele("copy-config")
        sub_ele(sub_ele(copy, "target"), ns.db)
        sub_ele(sub_ele(copy, "source"), "config").append(data.getroot())
        reply = mc.rpc(copy)
        return reply._root[0]


class EditConfig(Operation):
    option = "edit-config"
    name = "edit_config"
    nargs = "?"
    arg_completion = staticmethod(completions.filename_arg_completion)
    help = ("Takes a filename (or '-' for standard input) as"
            " argument. The contents of the file"
            " is data for a single NETCONF edit-config operation"
            " (put into the <config> XML element)."
            " Takes an optional --db argument, default is 'running'.")
    command_opts = ["db", "test"]

    def invoke(self, mc, ns, filename="-"):
        data = etree.parse(sys.stdin if filename == "-" else open(filename, "r"))
        config = new_ele("config")
        config.append(data.getroot())
        reply = mc.edit_config(config, test_option=ns.test, target=ns.db)
        return reply._root[0]


class PathExpressionException(Exception):
    pass


class XpathParser(object):
    """Simplified xpath expression parser building a XML tree.

    The supported xpath syntax is very limited:

    `/prefix:node[child1=value1][...]/prefix:node/...`

    All prefixes used need to be present in the namespace map. It is necessary
    to use prefixes only when crossing namespace boundary, otherwise the node
    inherits the namespaces from its parent.
    """

    QuotedT = r"%(q)s(?P<%(vname)s>[^%(q)s\\]*(?:\\.[^%(q)s\\]*)*)%(q)s"
    DirectValRx = r"(?P<value3>[^\]]*)"
    IdentT = "(?P<%(iname)s>[^:='\"\[/]+)"
    NodeSpecT = "(?:%s:)?%s"
    PredicateRx = ("\[ *%s(?: *= *(?:%s|%s|%s))? *\]" %
                   (NodeSpecT % (IdentT % {"iname": "pprefix"},
                                 IdentT % {"iname": "ptag"}),
                    QuotedT % {"q": "'", "vname": "value1"},
                    QuotedT % {"q": '"', "vname": "value2"},
                    DirectValRx))
    XpathRx = "(?:/%s(?:%s)*)+" % (NodeSpecT %
                                   (IdentT % {"iname": "prefix"},
                                    IdentT % {"iname": "tag"}),
                                   PredicateRx)
    XpathExpr = re.compile(XpathRx)
    NodeExpr = re.compile(NodeSpecT % (IdentT % {"iname": "prefix"},
                                       IdentT % {"iname": "tag"}))
    PredicateExpr = re.compile(PredicateRx)

    def __init__(self, expression, nsmap):
        self.expression = expression
        self.nsmap = nsmap

    def build_tree_on(self, top_node):
        "Build the path-matching tree for the expression rooted at given node."
        # Something better than doing that manually?
        if XpathParser.XpathExpr.match(self.expression) is None:
            raise PathExpressionException("Invalid path expression")
        self.pos = 0
        try:
            return self.build_tree(top_node)
        except Exception as e:
            raise PathExpressionException("Failed to parse path expression", e)

    def build_tree(self, element, namespace=None):
        "Build the tree for the remaining part of the expression."
        nmatch = XpathParser.NodeExpr.match(self.expression, self.pos + 1)
        gdict = nmatch.groupdict()
        (child, namespace) = self.build_node(element,
                                             gdict["tag"],
                                             gdict["prefix"],
                                             namespace)
        self.pos = nmatch.end()
        self.build_predicate(child, namespace)
        if self.pos < len(self.expression) and self.expression[self.pos] == "/":
            return self.build_tree(child, namespace)
        else:
            return (child, self.expression[self.pos:])

    def build_node(self, element, tag, pfx, namespace):
        elem_ns = self.nsmap.get(pfx, namespace)
        if elem_ns is None:
            expr = tag
        else:
            expr = "{%s}%s" % (elem_ns, tag)
        node = etree.SubElement(element, expr)
        return (node, elem_ns)

    def build_predicate(self, node, namespace):
        pmatch = XpathParser.PredicateExpr.match(self.expression, self.pos)
        if pmatch is None:
            return
        gdict = pmatch.groupdict()
        (child, nns) = self.build_node(node,
                                       gdict["ptag"], gdict["pprefix"],
                                       namespace)
        values = [gdict[valname]
                  for valname in ("value1", "value2", "value3")
                  if gdict[valname] is not None]
        if values != []:
            child.text = values[0]
        self.pos = pmatch.end()
        return self.build_predicate(node, namespace)


class ModifOp(Operation):
    def invoke(self, mc, ns, expr):
        pfxmap = nsmap(ns)
        config = new_ele("config", nsmap=pfxmap)
        (target, rest) = XpathParser(expr, pfxmap).build_tree_on(config)
        self.modify_target(ns, target, rest)
        reply = mc.edit_config(config, test_option=ns.test, target=ns.db)
        return reply._root[0]


class Set(ModifOp):
    option = name = "set"
    help = ("Takes an expression in the form `path=value`. The path may use "
            "prefixes defined with --ns.")
    nargs = 1
    command_opts = ["db", "test", "operation"]

    def modify_target(self, ns, target, rest):
        if rest == "" or rest[0] != "=":
            raise PathExpressionException(
                "The expression needs to take form <path>=<value>")
        attrib = "{%s}%s" % (ncclient.xml_.BASE_NS_1_0, "operation")
        target.attrib[attrib] = ns.operation
        target.text = rest[1:]


class Delete(ModifOp):
    option = name = "delete"
    help = ("Takes a path as an argument. The path may use "
            "prefixes defined with --ns.")
    nargs = 1
    command_opts = ["db", "test", "deloperation"]

    def modify_target(self, ns, target, rest):
        if rest != "":
            raise PathExpressionException(
                "The expression must be a valid path")
        attrib = "{%s}%s" % (ncclient.xml_.BASE_NS_1_0, "operation")
        target.attrib[attrib] = ns.deloperation


class Create(ModifOp):
    option = name = "create"
    help = ("Takes a path to a node to be created as an argument. "
            "The path may use prefixes defined with --ns.")
    nargs = 1
    command_opts = ["db", "test"]

    def modify_target(self, ns, target, rest):
        if rest != "":
            raise PathExpressionException(
                "The expression must be a valid path")
        attrib = "{%s}%s" % (ncclient.xml_.BASE_NS_1_0, "operation")
        target.attrib[attrib] = "create"


class GetSchema(Operation):
    option = "get-schema"
    name = "get_schema"
    help = "Takes an identifier (typically YANG module name) as parameter"
    nargs = 1

    def invoke(self, mc, ns, identifier):
        reply = mc.get_schema(identifier)
        return reply._root[0]


class NotificationListener(SessionListener):

    def callback(self, root, raw):
        tag, attrs = root
        if tag != qualify("notification", NOTIFICATION_NS):
            return
        print(raw)

    def errback(self, err):
        # ignore
        pass


class CreateSubscription(Operation):
    option = "create-subscription"
    name = "create_subscription"
    help = "Takes an optional stream name as a parameter, and an optional -x for XPath filtering"
    nargs = "?"
    command_opts = ["xpath"]

    listener = None

    def start_listener(self, mc):
        if self.listener is not None:
            return
        self.listener = NotificationListener()
        mc._session.add_listener(self.listener)

    def invoke(self, mc, ns, stream=None):
        sub = etree.Element(qualify("create-subscription", NOTIFICATION_NS))
        if stream is not None:
            etree.SubElement(sub, qualify("stream", NOTIFICATION_NS)).text = stream
        if ns.xpath is not None:
            etree.SubElement(sub, qualify("filter", NOTIFICATION_NS), type="xpath", select=ns.xpath)
        self.start_listener(mc)
        reply = mc.rpc(sub)
        return reply._root[0]


class Lock(Operation):
    name = option = "lock"
    help = "Lock the database."
    command_opts = ["db"]

    def invoke(self, mc, ns):
        reply = mc.lock(ns.db)
        return reply._root[0]


class Unlock(Operation):
    name = option = "unlock"
    help = "Unlock the database."
    command_opts = ["db"]

    def invoke(self, mc, ns):
        reply = mc.unlock(ns.db)
        return reply._root[0]


class Rpc(Operation):
    option = name = "rpc"
    help = ("Takes an optional filename (or '-' for standard input) as "
            " argument. The contents of the file"
            " is a single NETCONF rpc operation (w/o the surrounding"
            " <rpc>).")
    nargs = "?"
    arg_completion = staticmethod(completions.filename_arg_completion)

    def invoke(self, mc, ns, filename="-"):
        rpc = etree.parse(sys.stdin if filename == "-" else open(filename, "r"))
        reply = mc.rpc(rpc.getroot())
        return reply._root[0]


class Sleep(Operation):
    option = name = "sleep"
    nargs = 1

    def invoke(self, _mc, _ns, timeout):
        time.sleep(float(timeout))
        return new_ele("ok")


def run_rpc_dry():
    # debugging: replace the RPC._request method to return the rpc object
    class _Reply(object):
        def __init__(self, rpc, subelem):
            elem = new_ele("rpc", {"message-id": rpc._id})
            elem.append(subelem)
            self.data = elem
            self._root = [elem]

    def convert_query(rpc, op):
        return _Reply(rpc, op)
    RPC._request = convert_query


OPERATIONS = [Hello, Get, GetConfig, KillSession, DiscardChanges,
              Lock, Unlock, Commit, Validate, CopyRunningToStartup,
              CopyConfig, EditConfig, Set, Delete, Create, GetSchema,
              CreateSubscription, Rpc, Sleep]

OPERATION_OPTS = {op.option: op for op in OPERATIONS}


def xpath_filter(ns, path=None):
    filter = None
    if path is None:
        path = ns.xpath
    if path is not None:
        filter = ("xpath", path)
    return filter


def nsmap(ns):
    if ns.ns is not None:
        return dict(ns_assignment.split("=") for ns_assignment in ns.ns)
    else:
        return {}



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
import readline
import getpass
from ncclient import manager, transport
from lxml import etree
from ncclient.operations.rpc import RPCError
from contextlib import contextmanager
import six
import traceback
import re

from . import nctransport, completions, operations


class OperationArgAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        self.operation = kwargs["const"]
        super(OperationArgAction, self).__init__(*args, **kwargs)

    def __call__(self, parser, ns, values, option_string=None):
        if self.operation.nargs == "?":
            values = [] if values == self.operation else [values]
        ns.operations.append((self.operation, values))


quoted_rgx = r'"[^\\"]*(?:\\.[^\\"]*)*"|[^ ]+'


def parse_expr_args(expression):
    expr = " " + expression
    spaces = re.compile(" +")
    quoted = re.compile(quoted_rgx)
    while expr != "":
        spm = spaces.match(expr)
        if spm is None:
            raise ParserException("Syntax error before " + expr.split()[0])
        expm = quoted.match(expr[spm.end():])
        if expm is None:
            raise ParserException("Syntax error before " + expr[spm.end():].split()[0])
        ex = expm.group()
        if ex[0] == '"':
            yield re.sub(r"\\(.)", r"\1", ex[1:-1])
        else:
            yield ex
        expr = expr[spm.end() + expm.end():]


class ExpressionOperation(operations.Operation):
    def __init__(self, exprparser, process_errors=True):
        self.exprparser = exprparser
        self.process_errors = process_errors

    def invoke(self, mc, ns, expression):
        expression = expression.strip()
        if expression == '':
            return None
        try:
            ns = self.exprparser.parse_args(parse_expr_args(expression),
                                            namespace=argparse.Namespace(**vars(ns)))
        except ParserException as e:
            print(e, file=sys.stderr)
            return None
        if hasattr(ns, 'cmd_parser') and ns.cmd_parser is not None:
            ns.cmd_parser.print_help()
            return None
        try:
            if hasattr(ns, "value"):
                return ns.op.invoke(mc, ns, ns.value)
            else:
                return ns.op.invoke(mc, ns)
        except RPCError as e:
            if self.process_errors:
                return e.xml
            raise
        except Exception as e:
            if self.process_errors:
                report_exception(e, ns.debug)
                return None
            raise


def report_exception(exc, debug=False):
    if debug:
        traceback.print_exc(file=sys.stderr)
    else:
        msg = str(exc)
        if len(msg) > 100:
            msg = msg[:80] + "..."
        print('Internal error (%s: %s)' % (exc.__class__.__name__, msg))


class ParserException(Exception):
    pass


class SafeParser(argparse.ArgumentParser):
    def error(self, message):
        raise ParserException(message)


def expression_parser(parsercls=argparse.ArgumentParser, custom_help=False, **kwargs):
    exprparser = parsercls(prog='', **kwargs)
    exprs = exprparser.add_subparsers(help="Netconf commands")
    for opcls in operations.OPERATIONS:
        op = opcls()
        parserargs = dict(parents=[command_options_parser(op)], help=op.help)
        if custom_help:
            parserargs['add_help'] = False
        cmdparser = exprs.add_parser(op.option, **parserargs)
        if custom_help:
            cmdparser.add_argument("-h", "--help", action="store_const", const=cmdparser,
                                   dest="cmd_parser", help="show this help message")
        if op.nargs == 1:
            cmdparser.add_argument("value")
        elif op.nargs == "?":
            cmdparser.add_argument("value", nargs=op.nargs)
        cmdparser.set_defaults(op=op)
    return exprparser


command_option_args = {
    "style": (["-s", "--outputStyle"],
              dict(dest="style", default=[],
                   choices=["plain", "noaaa"], nargs="*")),
    "db": (["--db"],
           dict(dest="db", default="running",
                help="Database for commands that operate on a database.")),
    "timeout": (["--timeout"],
                dict(dest="timeout", default=600, type=int,
                     help="Confirmed commit timeout (in seconds).")),
    "wdefaults": (["--with-defaults"],
                  dict(dest="wdefaults", default=None,
                       choices=["explicit", "trim", "report-all", "report-all-tagged"],
                       help="Use with --get, --get-config, or --copy-config.")),
    "winactive": (["--with-inactive"],
                  dict(dest="winactive", action="store_true",
                       help="Send with-inactive parameter.  Use with --get, "
                       "--get-config, --copy-config, or --edit-config.")),  # FIXME: not supported
    "xpath": (["-x", "--xpath"],
              dict(dest="xpath",
                   help="XPath filter to be used with --get, --get-config, "
                   "and --create-subscription; kept for backward comptibility, "
                   "the operations accept direct arguments if provided.")),
    "test": (
        ["-t", "--test-option"],
        dict(dest="test", default="test-then-set",
             choices=["test-then-set", "set", "test-only"],
             help="Value of test-option used with edit-config (defaults to test-then-set)")),
    "operation": (
        ["-o", "--operation"],
        dict(dest="operation", default="merge",
             choices=["merge", "replace", "create"],
             help="Value of the operation attribute used with --set.")),
    "deloperation": (
        ["--del-operation"],
        dict(dest="deloperation", default="remove",
             choices=["remove", "delete"],
             help="Value of the operation attribute used with --delete."))
}


def command_options_parser(operation=None):
    cmd_opt_parser = argparse.ArgumentParser(add_help=False)
    cmd_group = cmd_opt_parser.add_argument_group("Command options")
    options = command_option_args.keys() if operation is None else operation.command_opts
    for opt in options:
        args, kws = command_option_args[opt]
        cmd_group.add_argument(*args, **kws)
    return cmd_opt_parser


def argparser():
    parser = argparse.ArgumentParser(prog="netconf-console", parents=[command_options_parser()])
    parser.add_argument("-v", "--version", dest="version",
                        help="force NETCONF version 1.0 or 1.1")  # FIXME: not supported
    parser.add_argument("-u", "--user", dest="username", default="admin",
                        help="username")
    parser.add_argument("-p", "--password", dest="password", default="admin", const=None, nargs="?",
                        help="password")
    parser.add_argument("--host", dest="host", default="127.0.0.1",
                        help="NETCONF agent hostname")
    parser.add_argument("--port", dest="port", default=2022, type=int,
                        help="NETCONF agent SSH port")
    parser.add_argument("--privKeyFile", help="File which contains the private key.")
    parser.add_argument("--raw", type=argparse.FileType("w"), nargs="?", const=sys.stdout)
    parser.add_argument("--tcp", action="store_true"),
    parser.add_argument("-N", "--ns", dest="ns",
                        help="Namespace prefix; useful for get queries with xpath filter.",
                        nargs="*")

    parser.add_argument("--debug", action="store_true")

    cmds = parser.add_argument_group("Commands")
    parser.set_defaults(operations=[])
    for opcls in operations.OPERATIONS:
        op = opcls()
        cmds.add_argument("--%s" % op.option,
                          dest=op.dest,
                          nargs=op.nargs,
                          action=OperationArgAction,
                          const=op,
                          choices=op.choices,
                          help=op.help)

    exprparser = expression_parser()
    parser.add_argument("-e", "--expr", action=OperationArgAction, nargs=1, const=ExpressionOperation(exprparser))

    parser.add_argument("--dry", action="store_true", default=False,
                        help="Do not send anything, return the RPC to be sent (debugging only).")

    parser.add_argument("--interactive", "-i", action="store_true",
                        help="Run the console interactively (do not use with commands).")
    return parser


class SSHSession(transport.SSHSession):
    def __init__(self, device_handler, raw_file):
        self.raw_file = raw_file
        super(SSHSession, self).__init__(device_handler)
        self.last_pos = 0

    def _parse10(self):
        with self.raw_processing(self._parsing_pos10):
            super(SSHSession, self)._parse10()

    def _parse11(self):
        with self.raw_processing(self._parsing_pos11):
            super(SSHSession, self)._parse11()

    @contextmanager
    def raw_processing(self, position):
        if self.raw_file is not None:
            self.raw_file.write(self._buffer.getvalue()[self.last_pos:].decode("utf8"))
        try:
            yield
        finally:
            self.last_pos = self._buffer.tell()


def connect(ns):
    password = ns.password
    if password is None:
        password = getpass.getpass()
    connect_args = dict(host=ns.host, port=ns.port,
                        username=ns.username, password=password,
                        key_filename=ns.privKeyFile,
                        hostkey_verify=False, look_for_keys=False, allow_agent=False)
    device_handler = manager.make_device_handler(None)
    if not ns.tcp and not ns.raw:
        # use the public API only
        session = transport.SSHSession(device_handler)
    elif ns.tcp:
        session = nctransport.TCPSession(device_handler, ns.raw)
    else:
        session = nctransport.SSHSession(device_handler, ns.raw)
    if not ns.tcp and ("hostkey_verify" not in connect_args or connect_args["hostkey_verify"]):
        session.load_known_hosts()
    try:
        session.connect(**connect_args)
    except Exception:
        if session.transport:
            session.close()
        raise
    return manager.Manager(session, device_handler, **connect_args)


def interactive_data():
    readline.parse_and_bind("tab: complete")
    readline.set_completer_delims(" ")
    readline.set_completer(completions.NCCompleter(operations.OPERATION_OPTS,
                                                   command_option_args))
    try:
        while True:
            command = six.moves.input('netconf> ')
            yield command
    except EOFError:
        return


def interactive_operations(exprparser, mc):
    if sys.stdin.isatty():
        data = interactive_data()
    else:
        data = sys.stdin
    operation = ExpressionOperation(exprparser)
    for command in data:
        if not mc.connected:
            return
        yield (operation, [command])
    return


def connect_and_process(ns):
    with connect(ns) as mc:
        try:
            if ns.interactive:
                expr_parser = expression_parser(SafeParser, custom_help=True)
                expr_parser.set_defaults(debug=ns.debug)
                operations_iter = interactive_operations(expr_parser, mc)
            else:
                operations_iter = ns.operations
            for (operation, args) in operations_iter:
                reply = operation.invoke(mc, ns, *args)
                if reply is not None:
                    print(etree.tostring(reply, encoding=six.text_type,
                                         pretty_print=("plain" not in ns.style)))
        except RPCError as e:
            print(etree.tostring(e.xml, encoding=six.text_type, pretty_print=True))
            return -1
    return 0


def main():
    parser = argparser()
    ns = parser.parse_args(sys.argv[1:])
    if ns.operations == []:
        ns.interactive = True
    if ns.interactive and ns.operations != []:
        parser.print_help()
        return 1
    if ns.dry:
        operations.run_rpc_dry()
    try:
        return connect_and_process(ns)
    except Exception as e:
        report_exception(e, ns.debug)
        return -1


if __name__ == "__main__":
    sys.exit(main())
